#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from ..ressource_classes import EditorAdaptor, adaptionmethod, \
    SingleResource, resource_property, resource_func
from types import GeneratorType
iter_types = (list,tuple,GeneratorType)



class ImageAdaptor(EditorAdaptor):
    
    pass
    


class ImageBase(ImageAdaptor.AdaptiveClass, SingleResource):
    
    
    
    def load(self, **kwargs):
        if self.ext in self.adaptor.multi_page_extensions:
            if "resolution" not in kwargs: kwargs["resolution"] = 300
        super().load(**kwargs)
        
        
    compression = resource_property("compression",
            jpg = ("jpeg", "j"),
            zip = "z")
    compr = compression
    
    compr_quality = resource_property("compr_quality")
    compression_quality = cquality = cqu = cqual = compr_quality
    
    resolution = resource_property("resolution")
    res = resolution
    
    size = resource_property("size")
    
    colortypenames = {"gray": ("grayscale", "g")}
    colortype = resource_property("colortype", **colortypenames)
    ctype = type = colortype
    
    colorspace = resource_property("colorspace", **colortypenames)
    cspace = color = colorspace
    
    @resource_func
    def resample(self,value_x, value_y=None, *args,**kwargs):
        if value_y is None: value_y = value_x
        return (value_x, value_y, *args), kwargs



# @ImageBase.make_multi_base
# class MultiImage(MultiResource):
#     # created as a subclass of ImageBase.Sequence (see above)
#
#     pass


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .... import launch, DependencyManager

with DependencyManager(__name__) as manager:
    manager.test_shellcommand("xdotool", pkg="xdotool")

names = (
    'BackSpace', 'Tab', 'Linefeed', 'Escape', 'space', 'exclam', 'quotedbl',
    'numbersign', 'dollar', 'percent', 'ampersand', 'apostrophe', 'parenleft',
    'parenright', 'asterisk', 'plus', 'comma', 'minus', 'period', 'slash',
    'colon', 'semicolon', 'less', 'equal', 'greater', 'question', 'at',
    'bracketleft', 'backslash', 'asciicircum', 'underscore', 'grave',
    'braceleft', 'bar', 'braceright', 'asciitilde', 'Delete', 'nobreakspace',
    'exclamdown', 'cent', 'sterling', 'currency', 'yen', 'brokenbar', 'section',
    'diaeresis', 'copyright', 'ordfeminine', 'guillemotleft', 'notsign',
    'hyphen', 'registered', 'macron', 'degree', 'plusminus', 'twosuperior',
    'threesuperior', 'acute', 'mu', 'paragraph', 'periodcentered', 'cedilla',
    'onesuperior', 'masculine', 'guillemotright', 'onequarter', 'onehalf',
    'threequarters', 'questiondown', 'Agrave', 'Aacute', 'Acircumflex',
    'Atilde', 'Adiaeresis', 'Aring', 'AE', 'Ccedilla', 'Egrave', 'Eacute',
    'Ecircumflex', 'Ediaeresis', 'Igrave', 'Iacute', 'Icircumflex',
    'Idiaeresis', 'ETH', 'Ntilde', 'Ograve', 'Oacute', 'Ocircumflex', 'Otilde',
    'Odiaeresis', 'multiply', 'Ooblique', 'Ugrave', 'Uacute', 'Ucircumflex',
    'Udiaeresis', 'Yacute', 'THORN', 'ssharp', 'agrave', 'aacute',
    'acircumflex', 'atilde', 'adiaeresis', 'aring', 'ae', 'ccedilla', 'egrave',
    'eacute', 'ecircumflex', 'ediaeresis', 'igrave', 'iacute', 'icircumflex',
    'idiaeresis', 'eth', 'ntilde', 'ograve', 'oacute', 'ocircumflex', 'otilde',
    'odiaeresis', 'division', 'oslash', 'ugrave', 'uacute', 'ucircumflex',
    'udiaeresis', 'yacute', 'thorn', 'ydiaeresis', 'Home', 'Insert', 'End',
    'Prior', 'Next', 'Help', 'Pause', 'VoidSymbol', 'Return', 'Break',
    'Caps_Lock', 'Num_Lock', 'Scroll_Lock', 'KP_Subtract', 'KP_End', 'KP_Add',
    'KP_Delete', 'KP_Right', 'KP_Insert', 'KP_Page_Up', 'KP_Divide',
    'KP_Page_Down', 'KP_Multiply', 'KP_Up', 'KP_Down', 'KP_Next', 'KP_Begin',
    'KP_Left', 'KP_Home', 'dead_grave', 'dead_acute', 'dead_circumflex',
    'dead_tilde', 'dead_diaeresis', 'dead_cedilla', 'Down', 'Left', 'Right',
    'Up', 'Shift', 'Control', 'Alt')


# def convert_kwargs(windowID=None, delay=5, clearmodifiers=False):
#     if windowID:
#         yield "--window"
#         yield windowID
#     if delay >= 0:
#         yield '--delay'
#         yield delay
#     if clearmodifiers:
#         yield "--clearmodifiers"

def text(character):
    return launch.get("xdotool", "type", "--delay", "0", character)

#
# def key(string: str, **kwargs):
#     #print("xdootol", string)
#     return launch.get("xdotool", "key", *convert_kwargs(**kwargs), string)

def press(string: str):
    return launch.get("xdotool", "keydown", "--delay", "0", string)


def rls(string: str):
    return launch.get("xdotool", "keyup", "--delay", "0", string)



#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#

from .. import DependencyManager

with DependencyManager(__name__) as tester:
    pystray = tester.import_module("pystray", pkg="pystray")
    Image = tester.import_module("PIL.Image", pkg="pillow")



import multiprocessing

def start(IconObject):
    IconObject.queue = multiprocessing.Queue()
    p = multiprocessing.Process(target=run_pystray_icon, args=(IconObject,))
    p.start()
    return p

def run_pystray_icon(IconObject):
    def get_putter(num):
        # we need a putter function that can be called without arguments for
        # each function to call when menu entry is chosen
        # functools.partial did somehow not work
        def putter():
            IconObject.queue.put(num)
        return putter
    menutuple = tuple(pystray.MenuItem(    IconObject.items[num][0],
            get_putter(num)    )     for num in range(len(IconObject.items)))
    menu = pystray.Menu(*menutuple)
    image = Image.open(IconObject.path)
    pystray.Icon(IconObject.tooltip, image, IconObject.tooltip, menu).run()
    
    



#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from ... import DependencyManager, launch

with DependencyManager(__name__) as manager:
    pyttsx3 = manager.import_module("pyttsx3")

engine = pyttsx3.init()


def call(text):
    engine.say(text)
    engine.runAndWait()
    
def set_volume(val):
    if val is None:
        return engine.getProperty('volume')
    engine.setProperty('volume', val)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from ... import DependencyManager

with DependencyManager(__name__) as manager:
    im_mod = manager.import_module("wand.image", pkg="wand")

Image = im_mod.Image

obj_class = Image
name_map = {"compr_quality": "compression_quality", "colortype":"type"}
allowed_values = {"colortype": im_mod.IMAGE_TYPES, "colorspace":
    im_mod.COLORSPACE_TYPES, "compression": im_mod.COMPRESSION_TYPES}

def load(file, **kwargs):
    with Image(filename=file, **kwargs) as im:
        s = im.sequence
        if len(s) == 0: raise ValueError
        if len(s) == 1: return im.clone()
        raise ValueError(f"More than one image in file {file}.")
    
def load_multi(file, **kwargs):
    with Image(filename=file, **kwargs) as im:
        s = im.sequence
        if len(s) == 0: raise ValueError
        return tuple(Image(i) for i in s)
    
multi_page_extensions = [".pdf", ".tif", ".tiff"]


def copy(obj):
    return obj.clone()
    

def construct_multi(sequence_of_objs):
    im  = Image()
    for o in sequence_of_objs:
        im.sequence.append(o)
    return im

def save(obj, destination):
    obj.save(filename = destination)
    
def close(obj):
    obj.destroy()




# def compression(obj, value):
#     if value is None: return obj.compression
#     obj.compression = value
#
# def compr_quality(obj, value):
#     if value is None: return obj.compression_quality
#     obj.compression_quality = value
#
# def resolution(obj, value):
#     if value is None: return obj.resolution
#     obj.resolution = value
#
# def size(obj, value):
#     if value is None: return obj.size
#     obj.size = value
#
# def colortype(obj, value):
#     if value is None: return obj.type
#     obj.type = value
#
# def resample(obj, value_x, value_y):
#     return obj.resample(value_x, value_y)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .. import launch, DependencyManager

with DependencyManager(__name__) as manager:
    manager.test_shellcommand("xsel", pkg="xsel")

xsel_names = {0:"b",1:"p",2:"s"}

def get(selection=0) -> str:
    try:
        return launch.get("xsel", "-o", "-" + xsel_names[selection])
    except launch.SubprocessError:
        return ""

def fill(content: str, selection=0):
    launch.get("xsel", "-i", "-" + xsel_names[selection], input=content)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#


__all__ = ["autoadapt", "keyb", "mouse", "ntfy", "dlg", "hook", "Icon", "Win",
    "sendwait", "nfsendwait", "clip", "Dfuncs", "events", "KeyWaiter",
    "TriggerManager", "Image", "mp3tag", "sleep", "sound","hotkeys","launch",
    "Layout","Dpowers", "Dhelpers"]

import os, sys
from time import sleep

dpowers_folder = os.path.dirname(os.path.realpath(__file__))
dpowers_startup_working_dir = os.getcwd()

try:
    import Dpowers  # this is to allow from Dpowers import * to also import the
                    # Dpowers variable itself
    import Dhelpers
except ModuleNotFoundError:
    sys.path.append(os.path.dirname(dpowers_folder))
    import Dpowers
    import Dhelpers
    

__version__ = "0.1.7rc1"
#Dpowers and Dhelpers share version number
if __version__ != Dhelpers.__version__:
    import warnings
    warnings.warn("Dpowers and Dhelpers packages do not share same version "\
                  "number")

from Dhelpers.adaptor import (AdaptorBase, adaptionmethod, AdaptionError,
    AdaptorBase, adaptionmethod, AdaptiveClass)
from Dhelpers.launcher import launch
from Dhelpers.builtin_manipulation import always_print_traceback, \
    restore_print_func
from Dhelpers.arghandling import ArgSaver
from Dhelpers.KeyboardLayouts import Layout

    
try:
    always_print_traceback()

    from . import default_backends

    class Adaptor(AdaptorBase):
        """Abstract baseclass for all of Dpower's Adaption classes."""
        #dependency_folder = dependency_folder
        NamedKeyClass = None  #set later
        NamedButtonClass = None  #set later
    
    Adaptor.set_default_backends(default_backends, evaluate_platform=True)
    
    DependencyManager = Adaptor.DependencyManager
    activate_autoadapt = Adaptor.activate_autoadapt
    adapt_all = Adaptor.adapt_all
    deactivate_autoadapt = Adaptor.deactivate_autoadapt
    check_adapted = Adaptor.check_adapted


    from .notificationpower import NotificationAdaptor
    ntfy = NotificationAdaptor(_primary_name="ntfy")


    from .windowpower import WindowAdaptor, WindowHandler
    class Win(WindowHandler):
        adaptor = WindowAdaptor(_primary_name="Win.adaptor")

    
    from .dialogpower import DialogAdaptor
    dlg = DialogAdaptor(_primary_name="dlg")


    from . import events
    from .events import keyb,mouse,hotkeys,hook, KeyWaiter, \
        KeyboardAdaptor, MouseAdaptor, HookAdaptor, SenderAdaptor, TriggerManager

    from .clipboardpower import ClipboardAdaptor
    clip = ClipboardAdaptor(_primary_name="clip")


    from .iconpower import IconAdaptor, IconHandler
    class Icon(IconHandler):
        adaptor = IconAdaptor(_primary_name="Icon.adaptor")
    
    from . import editing
    from .editing import Image, ImageAdaptor, mp3tag, mp3tagAdaptor
    
    from . import sound

    from . import Dfuncs
    from .Dfuncs import sendwait, nfsendwait
    
    from . import winapps

finally:
    restore_print_func()

#clean up this namespace
del os, AdaptorBase, adaptionmethod, always_print_traceback, \
    restore_print_func

Adaptor._set_effective_paths(globals(),__name__)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import functools


from .. import DependencyManager

with DependencyManager(__name__) as tester:
    tk = tester.import_module("tkinter", pkg="python3-tk", install_tool ="apt")

def popup(*buttons, xpos: int, ypos: int, x_rel_mouse: bool,y_rel_mouse: bool):
    """
    Function to show popup Menu with tkinter.
    _ can be used to specifiy hotkey characters.
    """
    
    root = tk.Tk()
    root.wm_attributes('-type', 'splash')
    root.focus_force()
    
    if x_rel_mouse:
        xpos = root.winfo_pointerx() - root.winfo_rootx() + xpos
        if xpos < 1: xpos = 1
    if y_rel_mouse:
        ypos = root.winfo_pointery() - root.winfo_rooty() + ypos
        if ypos < 1: ypos = 1
    root.geometry('+%s+%s'%(xpos, ypos))
    
    pressed_num = None
    
    def return_val(button_num, event=None):
        nonlocal pressed_num
        pressed_num = button_num
        root.destroy()
    
    root.bind_all("<FocusOut>", functools.partial(return_val, None))
    
    for num in range(len(buttons)):
        root.bind_all(str(num + 1), functools.partial(return_val, num))
        root.bind_all("<KP_%s>"%str(num + 1)
                ,functools.partial(return_val, num))
        
        i = str(num + 1) + " " + buttons[num]
        pos = i.find("_")
        if pos != -1 and pos + 1 < len(i):
            i = i[:pos] + i[pos + 1:]
            root.bind_all(i[pos].lower(), functools.partial(return_val, num))
        
        b = tk.Button(root, text=i, command=functools.partial(return_val, num),
                underline=pos, anchor="w")
        b.pack(fill="x")
    
    tk.mainloop()
    if pressed_num is not None:
        return buttons[pressed_num]



#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from Dhelpers.launcher import (launch, terminate_process, get_running_process)
import os, multiprocessing, warnings
from .. import ntfy, Dfuncs, Adaptor, adaptionmethod

class IconAdaptor(Adaptor):
    
    @adaptionmethod
    def start(self, IconObject):
        p = self.start.target_with_args()
        # p must be a pid (positive int) or Process object having a .pid
        # attribute. See the get_running_processes function definition
        # this will return a psutil.Process instance
        return get_running_process(p)
    

class IconHandler(IconAdaptor.AdaptiveClass):
    
    def __init__(self):
        super().__init__()
        self._items = []
        self.couple = True
        self.path = os.path.join(os.path.dirname(__file__),"icons","Dicon.png")
        self.tooltip = "Dpowers"
        self.left_click = self.terminate_parent
            #by default a left click terminates this script
        self.child_process = None
        self.last_item = ("Quit", self.terminate_parent)
        self.parent_id = os.getpid()
        self.queue = None
    
    def menuitem(self, text=None, func=None):
        """Adds a new entry to the icon's context menu.

        :param str text: Text of the menu entry. If omitted
            ``func.__name__`` will be used as text
            (underscores are replaced by space).
        :param func: Function to execute when clicking this menu entry.

        """
        
        if not func:
            def default_entry():
                warnings.warn("No action defined.")
            func = default_entry
        if not text: text = func.__name__.replace("_", " ")
        self._items.append((str(text), func))
    
    
    def additem(self, func):
        """A decorator to apply :func:`menuitem` directly onto a function."""
        self.menuitem(func=func)
        return func
    
    def add_default_menuitems(self):
        self.menuitem("Display win info", Dfuncs.display_win_info)
        self.menuitem("Get single key name(s)", Dfuncs.display_key_names)
        self.menuitem("Monitor input events for 10s",
                Dfuncs.monitor_input_events)
        self.menuitem("Toggle notifications", ntfy.toggle_all)
        self.menuitem("Save key stroke pattern", Dfuncs.save_key_replay)
        self.menuitem("Replay last key stroke pattern",
                Dfuncs.replay_pressed_keys)
    
    @property
    def items(self):
        if self.last_item:
            return self._items + [self.last_item]
        else:
            return self._items
    
    def start(self):
        self.child_process = self.adaptor.start(self)
        if self.couple:
            self.couple_process = launch.CoupleProcess(self.child_process)
        if self.queue is not None:
            if not isinstance(self.queue,multiprocessing.queues.Queue):
                raise ValueError
            # this happens if the start method defined a corresponding
            # queue object
            launch.thread(self.run_menu_funcs, initial_time_delay=1)
        
    
    def terminate(self):
        terminate_process(self.child_process)
    
    def terminate_parent(self):
        terminate_process(self.parent_id)
        
    def run_menu_funcs(self):
        while self.child_process.is_running():
            obj = self.queue.get()
            if isinstance(obj,int):
                func = self.items[obj][1]
            elif isinstance(obj,str):
                print(obj)
                for i in self.items:
                    if obj == i[0]:
                        # print("execute: "+str(i[1]))
                        func = i[1]
            else:
                raise TypeError
            launch.thread(func)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .. import Adaptor, adaptionmethod


class DialogAdaptor(Adaptor):
    @adaptionmethod
    def date(self, selected=None, **kwargs):
        """selected_date must be of form (dd,mm,yyyy). returns the same
        format."""
        return self.date.target_with_args()
    
    @adaptionmethod
    def path(self, selected=None, multi=False, dir=False, **kwargs):
        return self.path.target_with_args()
    
    @adaptionmethod
    def savepath(self, selected=None, overwrite=False, **kwargs):
        return self.savepath.target_with_args()
    
    @adaptionmethod
    def msg(self, text, title="Message dialog", error=False, warning=False,
            **kwargs):
        if title == "Message":
            if error:
                title = "Error dialog"
                if warning: raise ValueError
            elif warning:
                title = "Warning dialog"
        return self.msg.target(text, title, error, warning, **kwargs)
    
    
    @adaptionmethod
    def quest(self, text, title="Question dialog", **kwargs):
        return self.quest.target_with_args()
    
    
    @adaptionmethod
    def inp(self, text="Enter a new text:", title="Text input dialog"):
        return self.inp.target_with_args()
    
    
    @adaptionmethod
    def choose(self, options, default=None, text="Select an option:",
            title="Select dialog", multi=False, **kwargs):
        if not isinstance(default, (list, tuple)): default = (default,)
        if not multi and len(default) > 1: raise ValueError(
                "Only one default value allowed unless multi=True.")
        return self.choose.target(options, default, text, title, multi,
                **kwargs)

    
    @adaptionmethod
    def popup(self,*buttons, xpos: int = 2, ypos: int = -25, x_rel_mouse: bool\
        = True,y_rel_mouse: bool = True):
        return self.popup.target_with_args()


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .. import Adaptor, adaptionmethod
from .baseclasses import CallbackHook, KeyhookBase, ButtonhookBase

class HookAdaptor(Adaptor):
    
    @adaptionmethod("Keyhook")
    def keys(self, callback = False, timeout = 60, *, capture = False,
            reinject_func = None, priority: int  = 0, dedicated_thread=
            False,  press=True, release=True, allow_multipress=False,
            write_rls=True, **custom_kwargs):
        return self.keys.target_with_args()
    
    @adaptionmethod("Buttonhook")
    def buttons(self, callback = False, timeout = 60, *, capture = False,
            reinject_func = None, priority: int  = 0,  dedicated_thread =
            False, press=True, release=True, write_rls=True, **custom_kwargs):
        return self.buttons.target_with_args()

    @adaptionmethod("Cursorhook")
    def cursor(self, callback = False, timeout = 60, *, capture = False,
            reinject_func = None, priority: int  = 0, dedicated_thread =
            False, **custom_kwargs):
        return self.cursor.target_with_args()


    @adaptionmethod("Customhook")
    def custom(self, callback = False, timeout=60, *, capture=False,
            reinject_func=None, priority: int = 0, dedicated_thread = False,
            **custom_kwargs):
        return self.custom.target_with_args()
    
    

    @keys.target_modifier
    def _km(self, target):
        self.Keyhook_class = target
        target.NamedClass = self.NamedKeyClass
        target.update_active_dict()
        self.key_translation_dicts = target.name_translation_dicts
        return target

    def add_key_translation(self, dic):
        self.key_translation_dicts += [dic]
        self.Keyhook_class.update_active_dict()


    @buttons.target_modifier
    def _bm(self, target):
        self.Buttonhook_class = target
        target.NamedClass = self.NamedButtonClass
        target.update_active_dict()
        self.button_translation_dicts = target.name_translation_dicts
        return target

    def add_button_translation(self, dic):
        self.button_translation_dicts += [dic]
        self.Buttonhook_class.update_active_dict()
    
    
    def keysbuttons(self, keybfunc=False, timeout=60, *, mousefunc=False,
            allow_multipress = False, **hookkwargs):
        # returns a HookContainer instance
        if mousefunc is False: mousefunc = keybfunc
        return  self.keys(keybfunc, timeout, allow_multipress= allow_multipress,
                 **hookkwargs) + self.buttons(mousefunc, timeout, **hookkwargs)
    
    def keyboard_mouse(self, keybfunc=False, timeout=60, *, cursorfunc=False,
            mousefunc=False, **hookkwargs):
        if cursorfunc is False: cursorfunc = keybfunc
        return self.keysbuttons(keybfunc, timeout, mousefunc=mousefunc,
                **hookkwargs) + self.cursor(cursorfunc, timeout)
    
    
    
# from Dhelpers.baseclasses import AdditionContainer
#
# class HookAdaptorContainer(AdditionContainer):
#     basic_class = HookAdaptor
#
#     @functools.wraps(HookAdaptor.keys)
#     def keys(self,*args,**kwargs):
#         return sum(m.keys(*args,**kwargs) for m in self.members)
#
#     @functools.wraps(HookAdaptor.buttons¹23#)
#     def buttons(self,*args,**kwargs):
#         return sum(m.buttons(*args,**kwargs) for m in self.members)
#
#     @functools.wraps(HookAdaptor.cursor)
#     def cursor(self,*args,**kwargs):
#         return sum(m.cursor(*args,**kwargs) for m in self.members)
#
#     @functools.wraps(HookAdaptor.keysbuttons)
#     def keysbuttons(self,*args,**kwargs):
#         return sum(m.keysbuttons(*args,**kwargs) for m in self.members)
#
#     @functools.wraps(HookAdaptor.keyboard_mouse)
#     def keyboard_mouse(self,*args,**kwargs):
#         return sum(m.keyboard_mouse(*args,**kwargs) for m in self.members)
    
    
    


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#

from ... import DependencyManager

with DependencyManager(__name__) as manager:
    uinput = manager.import_module("evdev_prepared.uinput")

global_uinput = uinput.global_uinput

from evdev.ecodes import KEY, EV_KEY

names = {}
backend_layout = "evdev"
layout = "us"
use_shifted = True



def iterator(dic):
    for a, b in dic.items():
        if isinstance(b, (tuple, list)):
            for c in b: yield a,c
        else:
            yield a, b

for integer, name in iterator(KEY):
    n = name[4:]
    n=n.lower()
    names[n] = integer
    
global_uinput.start()

def press(keynumber):
    global_uinput.write(EV_KEY, keynumber, 1)

def rls(keynumber):
    global_uinput.write(EV_KEY, keynumber, 0)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import inspect

from .. import Adaptor, adaptionmethod
from Dhelpers.adaptor import AdaptionMethod
from types import GeneratorType
import os, functools, warnings
from types import FunctionType
from Dhelpers.file_iteration import FileIterator
from Dhelpers.baseclasses import iter_all_vars
from Dhelpers.named import NamedObj

path = os.path

iter_types = (list, tuple, GeneratorType)
class ClosedResource: pass
class ClosedResourceError(Exception): pass


edit_tag = "__edit"


class NamedValue(NamedObj):
     pass


class EditorAdaptor(Adaptor):
    
    property_value_dicts = dict()
    resource_class = None
    
    baseclass = True  #for Adaptor mechanism
    save_in_place = False  #if False, automatically rename before saving
    
    
    def _check(self, obj):
        if isinstance(obj, iter_types):
            for o in obj:
                if not isinstance(o, self.obj_class): raise TypeError
        else:
            if not isinstance(obj, self.obj_class): raise TypeError
        return obj
    
    @adaptionmethod(require=True)
    def load(self, file, **load_kwargs):
        backend_object = self._check(self.load.target_with_args())
        return backend_object
    
    obj_class = load.adaptive_property("obj_class", NotImplementedError)
    name_map = load.adaptive_property("name_map", {}, dict)
    allowed_values = load.adaptive_property("allowed_values", {}, dict)

    @load.target_modifier
    def _load_tm(self, target):
        self.create_effective_dicts()
        return target

    def create_effective_dicts(self, resource_class=None):
        if resource_class is None:
            resource_class = self.resource_class
            if not resource_class: return False
        assert resource_class.adaptor is self
        self.resource_class = resource_class
        property_value_dicts = dict()
        for propname in resource_class.list_resource_properties():
            prop = getattr(resource_class, propname)
            named_class = type(f"NamedValue_for_property_{propname}",
                  (NamedValue,), {})
            value_dict = prop.fget._value_dict
            for stand_name, names in value_dict.items():
                if isinstance(names,str): names = (names,)
                named_val = named_class(stand_name, *names)
                    #this registered the defined names as a new NamedObj inst
            stand_dict = named_class.StandardizingDict()
            try:
                value_list = self.allowed_values[propname]
            except KeyError:
                pass
            else:
                assert isinstance(value_list, (list,tuple))
                stand_dict.update(value_list)
                stand_dict.other_dict.clear() #this only contains mappings on
                # itself anyway
            property_value_dicts[propname] = stand_dict
        self.property_value_dicts = property_value_dicts

    @adaptionmethod
    def load_multi(self, file, **load_kwargs):
        backend_objects = self._check(self.load_multi.target_with_args())
        return backend_objects
    
    multi_page_extensions = load_multi.adaptive_property(
             "multi_page_extensions", [])
    
    @adaptionmethod
    def construct_multi(self, sequence_of_backend_objs):
        sequ = tuple(sequence_of_backend_objs)
        self._check(sequ)
        return self.construct_multi.target(sequ)
        
    @adaptionmethod(require=True)
    def save(self, backend_obj, destination=None):
        self._check(backend_obj)
        if self.save_in_place:
            assert destination is None
        else:
            assert  destination is not None
        self.save.target(backend_obj, destination)
    
    @adaptionmethod(require=True)
    def close(self, backend_obj):
        self._check(backend_obj)
        self.close.target_with_args()
        
    @adaptionmethod
    def copy(self, backend_obj):
        self._check(backend_obj)
        return self._check(self.copy.target_with_args())
    
    def _check_method(self, name):
        method = None
        try:
            amethod = getattr(self,name)
        except AttributeError:
            pass
        else:
            if isinstance(amethod,AdaptionMethod): method = amethod
        if method is None:
            target_space = self.load.target_space
            try:
                method = getattr(target_space, name)
            except AttributeError:
                return False, False
        if not callable(method): return False, False
        argnum = len(inspect.signature(method).parameters)
        if argnum < 2: raise TypeError
        return method, argnum
    
    def forward_property(self, name, backend_obj, value = None):
        stand_dict = self.property_value_dicts.get(name,{})
        method, argnum = self._check_method(name)
        if method:
            assert argnum == 2
            return method(backend_obj, value)
        backend_name = self.name_map.get(name, name)
        if value is None: #in this case the property is retrieved
            ret = getattr(backend_obj, backend_name)
            if isinstance(ret,str) and stand_dict:
                try:
                    ret = stand_dict.NamedClass.get_stnd_name(ret)
                except KeyError:
                    pass
            return ret
        value = stand_dict.get(value,value)
        setattr(backend_obj, backend_name, value)
    
    
    def forward_func_call(self, name, backend_obj, *args, **kwargs):
        method, argnum = self._check_method(name)
        if method: return method(backend_obj, *args,**kwargs)
        backend_name = self.name_map.get(name, name)
        return getattr(backend_obj, backend_name)(*args,**kwargs)
    
    
    # def _inspect_unkown(self, name):
    #     backend_name = self.name_map.get(name, name)
    #     obj = getattr(self.obj_class, backend_name)
    #         # this might raise AttributeError!
    #     if callable(obj):
    #         attr_type = 1
    #     elif isinstance(obj, property):
    #         attr_type = 2
    #     else:
    #         raise AttributeError(obj)
    #     return backend_name, attr_type
    #
    # def get_unknown(self, name, backend_obj):
    #     method, argnum = self._check_method(name)
    #     if method:
    #         if argnum == 2: return method(backend_obj, None)
    #         return functools.partial(method,backend_obj)
    #     backend_name, attr_type = self._inspect_unkown(name)
    #     return getattr(backend_obj, backend_name)
    #         #this might raise Attribute error
    #         # otherwise this is a callable object, or the result of a
    #         # property
    #
    # def set_unknown(self, name, backend_obj, value):
    #     method, argnum = self._check_method(name)
    #     if method:
    #         method(backend_obj, value)
    #         return True
    #     try:
    #         backend_name, attr_type = self._inspect_unkown(name)
    #     except AttributeError:
    #         return
    #     if attr_type is 1: return #ignore forwarding if its a callable
    #     setattr(backend_obj, backend_name , value) #only happens for properties
    #     return True
        
        
# def standardize(val, value_dict):
#     if value_dict:
#         for stan_name, used_names in value_dict.items():
#             if val == used_names or val in used_names: val = stan_name
#     return val

def resource_property(name, **value_dict):
    def func(self):
        return self.forward_property(name)
    func.__name__ = name
    func.__qualname__ = name
    func._resource = True
    func._value_dict = value_dict
    func = property(func)
    @func.setter
    def func(self, val):
        self.forward_property(name,val)
    return func

# def multi_resource_property(name):
#     func = lambda self: self.get_from_single(name)
#     func.__name__ = name
#     func.__qualname__ = name
#     func._resource = True
#     func = property(func)
#     @func.setter
#     def func(self, val):
#         for single in self.sequence: setattr(single,name,val)
#     return func

def is_resource_property(obj):
    return isinstance(obj, property) and hasattr(obj.fget,"_resource")



def resource_func(name_or_func, *names):
    """A decorator to create resource_funcs"""
    if isinstance(name_or_func, FunctionType):
        inp_func = name_or_func
        name = inp_func.__name__
        def func(self, *args, **kwargs):
            args, kwargs = inp_func(self,*args,**kwargs)
            return self.forward_func_call(name, *args, **kwargs)
        func.__name__ = name
        func.__qualname__ = name
        func._resource = True
        func._names = names
        func._inp_func = inp_func
        return func
    if isinstance(name_or_func, str):
        def decorator(inp_func):
            return resource_func(inp_func, name_or_func,*names)
        return decorator
    raise TypeError
#
#
# def multi_resource_func(name):
#     def func(self,*args,**kwargs):
#         return self.get_from_single(name,True,*args,**kwargs)
#     func.__name__ = name
#     func.__qualname__ = name
#     func._resource = True
#     return func

def is_resource_func(obj):
    return callable(obj) and hasattr(obj,"_resource") and obj._resource is True





class Resource:
    
    file = None
    adaptor = None  #inherited as AdaptiveClass subclass

    allowed_file_extensions = None
    attr_redirections = None
    
    @classmethod
    def list_resource_properties(cls):
        for name in iter_all_vars(cls,is_resource_property):
            realname = getattr(cls,name).fget.__name__
            if name == realname: yield name
    
    @classmethod
    def list_resource_funcs(cls):
        for name in iter_all_vars(cls, is_resource_func):
            realname = getattr(cls, name).__name__
            if name == realname: yield name

    # @classmethod
    # def _update_redirections(cls):
    #     cls.attr_redirections = {}
    #     for name in cls.list_resource_properties():
    #         cls.attr_redirections[name] = name
    #         names = getattr(cls,name).fget._names
    #         for n in names: cls.attr_redirections[n] = name
    #     for name in cls.list_resource_funcs():
    #         cls.attr_redirections[name] = name
    #         names = getattr(cls,name)._names
    #         for n in names: cls.attr_redirections[n] = name
    #     return cls.attr_redirections

    def filepath_split(self):
        if self.file is None: return
        folder, filename = path.split(self.file)
        name, ext = path.splitext(filename)
        self.folder = folder
        self.filename = name
        self.ext = ext.lower()

    def get_destination(self, input_dest=None, num=None, insert_folder=False):
        ext = None
        inp_name = None
        if input_dest:
            if input_dest.startswith("."):
                inp_name, ext = "", input_dest
            else:
                inp_name, ext = path.splitext(input_dest)
        elif self.adaptor.save_in_place:
            return
        if not ext: ext = self.ext
        if insert_folder:
            folder_name = insert_folder if isinstance(insert_folder,
                    str) else self.filename + edit_tag
        if not inp_name:
            if insert_folder:
                folder = path.join(self.folder, folder_name)
                os.makedirs(folder, exist_ok=True)
                append = f"_{num}" if num else ""
            else:
                folder = self.folder
                append = f"_{num}" if num else edit_tag
            return path.join(folder, self.filename + append + ext)
        if insert_folder:
            if not path.split(inp_name)[0]:
                # that means inp_name is a file not a folder structure
                os.makedirs(folder_name, exist_ok=True)
                inp_name = path.join(folder_name, inp_name)
        append = f"_{num}" if num else ""
        try:
            pages = self.pages
        except AttributeError:
            pass
        else:
            if pages and ext not in self.adaptor.multi_page_extensions:
                raise ValueError
        return inp_name + append + ext

    def _save(self, dest=None):
        raise NotImplementedError

    def save(self, destination=None, num=None):
        dest = self.get_destination(destination, num)
        return self._save(dest)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        
    def close(self):
        raise NotImplementedError
    
    
class SingleResource(Resource):
    
    multi_base = None
    _backend_obj = None
    #_inst_attributes = "file", "sequence", "backend_obj"
    
    # @classmethod
    # def make_multi_base(cls, MultiClass):
    #     """A decorator"""
    #     cls.multi_base = MultiClass
    #     cls._set_multi(MultiClass)
    #     return MultiClass
    
    def __init_subclass__(cls):
        super().__init_subclass__()
        try:
            func = cls.adaptor.create_effective_dicts
        except AttributeError:
            pass
        else:
            func(cls)
        # cls._update_redirections()
        # cls._set_multi(cls.multi_base)
    
    

    # @classmethod
    # def _set_multi(cls, multi_base):
    #     class multi(multi_base):
    #         __qualname__ = f"{cls.__name__}.multi"
    #         __name__ = __qualname__
    #         __module__ = cls.__module__
    #         SingleClass = cls
    #         attr_redirections = cls.attr_redirections
    #     for name in cls.list_resource_properties():
    #         rp = multi_resource_property(name)
    #         setattr(multi,name,rp)
    #     for name in cls.list_resource_funcs():
    #         rf = multi_resource_func(name)
    #         setattr(multi,name,rf)
    #     cls.multi = multi
        

    def __init__(self, file=None, backend_obj=None, **load_kwargs):
        if isinstance(file, iter_types) or isinstance(backend_obj, iter_types):
            raise ValueError
        self.file = file
        self.backend_obj = backend_obj
        self.pages = None
        if file:
            assert backend_obj is None
            self.filepath_split()
            self.load(**load_kwargs)
        elif backend_obj:
            self.adaptor._check(backend_obj)
            
    def load(self, **kwargs):
        #print(self.ext, self.adaptor.multi_page_extensions)
        if self.ext in self.adaptor.multi_page_extensions:
            backend_objs = self.adaptor.load_multi(self.file, **kwargs)
            self.pages = tuple(
                    self.__class__(backend_obj=bo) for bo in backend_objs)
        else:
            self.backend_obj = self.adaptor.load(self.file, **kwargs)
        
        
    @classmethod
    def iterator(cls, *args,**kwargs):
        for filepath in FileIterator(*args,**kwargs).generate_paths():
            yield cls(filepath)

    @property
    def backend_obj(self):
        if self._backend_obj is ClosedResource: raise ClosedResourceError
        return self._backend_obj

    @backend_obj.setter
    def backend_obj(self, value):
        self._backend_obj = value
    
    
    def backend_objs(self):
        return tuple(im.backend_obj for im in self.pages)
    
    def construct_backend_obj(self):
        return self.adaptor.construct_multi(self.backend_objs())
        
    def _save(self, dest=None):
        if self.pages:
            self.adaptor.save(self.construct_backend_obj(), dest)
        else:
            self.adaptor.save(self.backend_obj, dest)
        return dest
    
    
    def close(self):
        """Close the resource"""
        if self.backend_obj: self.adaptor.close(self.backend_obj)
        if self.pages:
            for resource in self.pages: resource.close()
        self.backend_obj = ClosedResource
        self.pages = ClosedResource
        
    def copy(self):
        if self.pages:
            new_inst = self.__class__()
            new_inst.pages = [im.copy() for im in self.pages]
        else:
            new_backend_obj = self.adaptor.copy(self.backend_obj)
            new_inst = self.__class__(backend_obj=new_backend_obj)
        new_inst.file = self.file
        new_inst.filepath_split()
        return new_inst
    
    
    def forward_property(self, name, val=None):
        if self.pages:
            rets = [p.forward_property(name,val) for p in self.pages]
            if val is None:
                ret = rets[0]
                if all(i == ret for i in rets): return ret
                return rets
        else:
            return self.adaptor.forward_property(name,self.backend_obj, val)
        
    def forward_func_call(self, name, *args, **kwargs):
        if self.pages:
            rets = [p.forward_func_call(name, *args, **kwargs) for p in
                self.pages]
            ret = rets[0]
            if all(i == ret for i in rets): return ret
            return rets
        else:
            return self.adaptor.forward_func_call(name, self.backend_obj,
                    *args,**kwargs)
    
    
    def __add__(self, other):
        if not isinstance(other,self.__class__):
            raise NotImplementedError
        assert self.adaptor == other.adaptor
        assert self.pages and other.pages
        new_inst = self.__class__()
        new_inst.pages = self.pages + other.pages
        return new_inst
    
    def __getitem__(self, item):
        if not self.pages: raise ValueError
        return self.pages[item]
    
    def __setitem__(self, key, value):
        if not self.pages: raise ValueError
        assert isinstance(value, self.__class__)
        assert value.adaptor == self.adaptor
        assert not value.pages
        self.pages[key] = value
        
    
    #
    # def __setattr__(self, key, value):
    #     if key not in self._inst_attributes:
    #         try:
    #             key = self.attr_redirections[key]
    #         except KeyError:
    #             if self.adaptor.set_unknown(key,self.backend_obj,value):
    #                 _print(f"WARNING: Attribute '{key}' used from backend "
    #                  f"'{self.adaptor.backend}'. Undefined in " f"Dpowers.")
    #                 return
    #     super().__setattr__(key, value)



# @SingleResource.make_multi_base
# class MultiResource(Resource):
#
#     SingleClass = None  # set by init subclass form Image class
#     _inst_attributes = "file", "sequence", "backend_obj"
#
#     def save(self, destination=None, split=False):
#         if destination and not destination.endswith(tuple(self.allowed_file_extensions)):
#             split = True
#         if not split: return super().save(destination)
#         l = len(self.sequence)
#         if l == 0: raise ValueError
#         first = self.sequence[0]
#         if l == 1:
#             if destination is None and self.adaptor.save_in_place:
#                 return first._save()
#             try:
#                 dest = self.get_destination(destination)
#             except AttributeError:
#                 try:
#                     dest = first.get_destination(destination)
#                 except AttributeError:
#                     raise ValueError("Could not find destination name.")
#             return first._save(dest)
#         destinations = []
#         for num in range(l):
#             single = self.sequence[num]
#             if destination is None:
#                 if self.adaptor.save_in_place: single._save()
#             try:
#                 dest = self.get_destination(destination, num=num+1,
#                         insert_folder=True)
#             except AttributeError:
#                 try:
#                     dest = single.get_destination(destination)
#                 except AttributeError:
#                     raise ValueError("Could not find destination name.")
#             destinations.append(single._save(dest))
#         return destinations
#
#
#     def get_from_single(self, name, func = False, *args, **kwargs):
#         ret = []
#         same = True
#         for single in self.sequence:
#             this = getattr(single,name)
#             if func: this = this(*args, **kwargs)
#             ret.append(this)
#             if this is not ret[0]: same=False
#         if same: return ret[0]
#         return ret
#
#
#     def __getattr__(self, key):
#         if key not in self._inst_attributes:
#             try:
#                 ret = self.get_from_single(key)
#             except AttributeError:
#                 pass
#             else:
#                 return ret
#         raise AttributeError(key)
    


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from Dhelpers.named import NamedObj
from .event_classes import Buttonevent, Keyvent, NamedPressReleaseEvent
import re

class NamedKeyButton(NamedObj):
    Event = None
    #__slots__ = ["press_event", "release_event", "release_event_without_rls"]
    
    def __init_subclass__(cls, event_baseclass="inherit"):
        super().__init_subclass__()
        if event_baseclass == "inherit": event_baseclass = cls.Event
            #inherit Event's baseclass from parent's event baseclass
        if not issubclass(event_baseclass, NamedPressReleaseEvent):
            raise TypeError
        class Event(event_baseclass):
            __qualname__ = f"{cls.__name__}.Event"
            __name__ = __qualname__
            __module__ = cls.__module__
            NamedClass = cls
        cls.Event = Event
    
    def __init__(self, *names):
        super().__init__(*names)
        # create all the corresponding events already to save time later
        self.press_event = self.Event(self.name, press=True)
        self.release_event = self.Event(self.name, press=False)
        self.release_event_without_rls = self.Event(self.name,
                press=False, write_rls=False)
    
    def get_event(self, press, write_rls=True):
        if press: return self.press_event
        if not write_rls:
            if press: raise ValueError
            return self.release_event_without_rls
        return self.release_event
    
    
    @classmethod
    def from_string(cls, string):
        event = 0
        for s in string.split(" "):
            if s.endswith("_rls"):
                e = cls.Event(s[:-4], press=False, NamedClass=cls)
            else:
                e = cls.Event(s, press=True, NamedClass=cls)
            event += e
        return event
    
    
    @classmethod
    def _check_if_name_is_allowed(cls, name):
        if isinstance(name, str) and len(name) > 1:
            for s in (" ", "+", "[", "]", "_rls", "<", ">"):
                if s in name: raise NameError(
                        "Keyname: " + str(name) + "\nKey class: " + str(
                                cls) + "\nThe string %s is not allowed."%s)
        return True
    
    @classmethod
    def analyze_string(cls, string, applied_func):
        if len(string) <= 1:
            return applied_func(string)
        splitted_name = re.split("([ +])", string)  # splits at + and space
        if len(splitted_name) == 1:
            return applied_func(string)
        output = ""
        for item in splitted_name:
            add_later = ""
            if item in (" ", "+", ""):
                output += item
                continue
            if item.endswith("_rls"):
                item = item[:-4]
                add_later = "_rls"
            if item.startswith("[") and item.endswith("]"):
                number = item[1:-1]
                if not number.isdecimal():
                    raise TypeError(
                            "Non decimal string enclosed in []:   " + number)
                output += "[%s]"%applied_func(number)
            else:
                output += applied_func(item)
            output += add_later
        return output



class NamedKey(NamedKeyButton, event_baseclass=Keyvent):
    #__slots__ = []
    pass

class NamedButton(NamedKeyButton, event_baseclass=Buttonevent):
    
    #__slots__ = []
    
    def get_event(self, press, write_rls=True, x=None, y=None):
        event = super().get_event(press, write_rls)
        event.x=x
        event.y=y
        
        



#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .. import Adaptor, adaptionmethod, ntfy, keyb
import time


class ClipboardAdaptor(Adaptor):
    selection_names = {
        0: "clipboard", 1: "primary selection", 2: "secondary selection"
        }
    
    @adaptionmethod
    def get(self, selection=0) -> str:
        """Retrieves the content
        
        :param selection: The selection to be retrieved. Defaults to 0,
            i.e. the standard clipboard.
        :returns str: retrieved content
        """
        return self.get.target_with_args()
    
    @adaptionmethod
    def fill(self, content:str, selection=0, notify=False):
        content=str(content)
        self.fill.target(content,selection)
        # now we check that the content was successfully pasted into the
        # selection:
        after = self.get(selection)
        if content.endswith("\n") and not after.endswith("\n"):
            after += "\n"
            # this is necessary because the use of launch.get in the xsel and
            #  xclip adaptions, will remove the last "\n" in the returned
            # after value in this case
        if after != content:
            raise ValueError("Failed to set clipboard content.\nExpected: '{"
                             "content}'\nActually: '{after}'".format_map(
                    locals()))
        if notify:
            ntfy("Saved to " + self.selection_names[selection], 3,after)
    
    
    def remove(self, selection=0):
        s = self.get(selection)
        self.fill("",selection)
        return s
    
    def monitor(self, selection=0, duration=60):
        for i in range(duration):
            ntfy("Monitoring " + self.selection_names[selection], 1,
                    self.get(selection))
            time.sleep(1)
    
    def saved(self,selection=0):
        return SelectionSaver(self,selection)
    
    def paste(self, string):
        with self.saved():
            self.fill(string)
            keyb.comb("ctrl", "v")



class SelectionSaver:
    """
    Saves the value of the clipboard inside a with clause is executed,
    and restores it automatically afterwards
    Usage:
    with saved() as s:
        some code(s)
        more code
    """
    
    def __init__(self, clipboardadaptor, selection=0):
        self.clipboardadaptor = clipboardadaptor
        self.selection = selection
    
    def __enter__(self):
        self.save = self.clipboardadaptor.remove(self.selection)
        return self.save
    
    def __exit__(self, *ignore_these_args):
        time.sleep(0.2)
        self.clipboardadaptor.fill(self.save, self.selection)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from ... import DependencyManager

with DependencyManager(__name__) as tester:
    uinput = tester.import_module("evdev_prepared.uinput")
    device_control = tester.import_module("evdev_prepared.device_control")

global_uinput = uinput.global_uinput
DeviceSelector = device_control.DeviceSelector
    
global_uinput.start()
names = {}

def select_device(**device_kwargs):
    selector = DeviceSelector(**device_kwargs)
    devs = selector.matching_devs()
    global_uinput.update_from_devices(*devs)
    for name, nums in iterator(devs):
        name = name.lower()
        names[name] = nums
    return devs

def iterator(devs):
    for dev in devs:
        caps = dev.capabilities(verbose=True)
        for event_type,event_list in caps.items():
            if len(event_type) != 2: raise ValueError(event_type)
            ev_type_name = event_type[0]
            ev_type_num = event_type[1]
            if ev_type_name == "EV_SYN": continue
            for entry in event_list:
                if len(entry) != 2: raise ValueError(entry)
                event_num = entry[1]
                nums = ev_type_num, event_num
                name = entry[0]
                if isinstance(name, (tuple, list)):
                    for n in name: yield n, nums
                else:
                    yield name, nums
    
def press(nums):
    ev_type_num, event_num = nums
    global_uinput.write(ev_type_num, event_num, 1)

def rls(nums):
    ev_type_num, event_num = nums
    global_uinput.write(ev_type_num, event_num, 0)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import types
from Dhelpers.launcher import launch, Application
from . import Win
from .windowpower import WindowSearch

@WindowSearch.register
class WindowApplication(Application):
    
    Win = Win
    mouse=NotImplemented
    
    def __init__(self, command=NotImplemented, check_installed=False):
        super().__init__(command, check_installed)
        self._winsearch = 0
        self.last_found = None


    @property
    def winsearch(self):
        if self.used_inst: return self.used_inst.winsearch
        return self._winsearch
        
        
    def add_winprops(self, *winargs,**winkwargs):
        if self.used_inst:
            self._winsearch = self.used_inst.winsearch
            self.used_inst = None
        if winargs and isinstance(winargs[0], (self.__class__, self.Win.Search)):
                assert not winkwargs
                for obj in winargs:
                    if isinstance(obj,self.__class__): obj = obj.winsearch
                    self._winsearch += obj
                return
        if "visible" not in winkwargs: winkwargs["visible"] = True
        self._winsearch += self.Win.Search(*winargs,**winkwargs)

    def __add__(self, other):
        assert isinstance(other,(self.__class__, self.Win.Search))
        combined = self.__class__()
        combined.add_winprops(self)
        combined.add_winprops(other)
        return combined
        
    
    def __getattr__(self, item):
        winobj = self.last_found.winsearch_object if self.last_found else \
            self.winsearch
        if winobj != 0:
            obj = getattr(winobj, item)
            if isinstance(obj,(types.FunctionType, types.MethodType)):
                return obj
            raise AttributeError
        
    def find(self):
        found_win = self.winsearch.find()
        if len(found_win) == 1: self.last_found = found_win
        return found_win
        
    find_win = find
    
    def startwait(self, *args, start=True, timeout=10,**kwargs):
        return self.wait(*args,start=start,timeout=timeout, **kwargs)
        
    def wait(self, *args,start=False,timeout=10,**kwargs):
        self._found = self.find()
        if start:
            cmd = start if isinstance(start, str) else self.command
            launch(cmd,*args, **kwargs)
        # capturing the pid of this process doesnt help, as windows might
        # share same pid of the one that was first opened
        new_win = self._found.wait_num_change(+1, timeout=timeout)
        if not new_win: return
        new_win.activate()
        self.last_found = new_win
        return new_win
        
    
    def center_mouse(self):
        geom = self.geometry()
        self.mouse.moveto(geom[0]+geom[2]/2, geom[1]+geom[3]/4)
        
        
class EditorApp(WindowApplication):
    
    def jump_to_line(self, file, line=None):
        args = []
        if line: args += ["--line", str(line)]
        args += [str(file)]
        launch(self.command, *args)
        self.wait_exist_activate()



pythoneditor = EditorApp()  #this instance is used in trigman.py
# the actual windows behind it, need to be added via .add_winprops method
# later on


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from Dhelpers.baseclasses import AdditionContainer
from Dhelpers.named import NamedObj
from . import hotkeys
from abc import ABC, abstractmethod
from time import sleep




class Event(AdditionContainer.Addend):
    
    device = None
    source = None  #should be a class or module
    
    def __repr__(self):
        return f"<Event '{self}' of subclass " \
            f"{self.__class__.__module__}.{self.__class__.__name__}>"

    def sending_version(self):
        return self
    
    def hotkey_version(self, rls=False):
        return self



class EventSequence(AdditionContainer, Event, basic_class=Event):
    def __str__(self):
        return StringEvent.sequence_symbol.join(str(m) for m in self.members)
    
    def sending_version(self):
        return self.__class__(*tuple(m.sending_version() for m in self.members))
    
    def hotkey_version(self, rls=False):
        return self.__class__(
                *tuple(m.hotkey_version(rls=rls) for m in self.members))


class EventCombination(Event):
    
    @property
    def members(self):
        return self._members
    
    def __init__(self, *args):
        if len(args) < 2: raise ValueError
        for a in args:
            if not isinstance(a, PressReleaseEvent): raise TypeError
            if a.press is False: raise ValueError
        self._members = args
    
    def __str__(self):
        return StringEvent.combination_symbol.join(str(m) for m in self.members)
    
    def sending_version(self):
        return sum(m for m in self.members) + sum(
                m.reverse() for m in reversed(self.members))
    
    def hotkey_version(self, rls=False):
        ret = sum(m for m in self.members)
        if rls: ret += self.members[-1].reverse()
        return ret




class MouseMoveEvent(Event):
    
    def __init__(self, x, y, relative = False, screen_coordinates=True):
        self.x = x
        self.y = y
        self.relative = relative
        self.abbr = "r" if relative else "a"
    
    def __str__(self):
        return f"(mouse_move,{self.abbr},{self.x},{self.y})"
        



class StringAnalyzeUtilitites(ABC):
    
    combination_symbol = "+"
    sequence_symbol = " "
    
    @classmethod
    def split_str(cls,string):
        splitted_string = string.split(cls.sequence_symbol)  # split at space
        # print("splitted", splitted_string)
        corrected_split = []
        skip_next = False
        for n in range(len(splitted_string)):
            if skip_next:
                skip_next = False
                continue
            part = splitted_string[n]
            if part.startswith(cls.combination_symbol):
                corrected_split[-1] += part
            else:
                corrected_split.append(part)
            if part.endswith(cls.combination_symbol):
                corrected_split[-1] += splitted_string[n + 1]
                skip_next = True  # print("corrected",corrected_split)
        # now the corrected split has taken the combinations into account
        for s in corrected_split:
            if cls.combination_symbol in s:
                yield s.split(cls.combination_symbol)
            else:
                yield s
    
    def create_from_str(cls_or_self, string, hotkey=False, **kwargs):
        events = []
        for entry in cls_or_self.split_str(string):
            if isinstance(entry, str):
                event = cls_or_self._create_from_str(entry, **kwargs)
                events.append(event)
                if hotkey:
                    try:
                        p = event.press
                    except AttributeError:
                        raise ValueError("Option hotkey=True "
                                         f"does not make sense for f{event}.")
                    if not p: raise ValueError(
                            "Hotkey mode enabled. Use option  hotkey=False to "
                            "only send press or release events.")
                    events.append(event.reverse())
            elif isinstance(entry, (tuple, list)):
                t = tuple(cls_or_self._create_from_str(item, **kwargs)  for
                        item in entry)
                events.append(EventCombination(*t))
            else:
                raise TypeError
        if len(events) == 0: return ""
        if len(events) == 1: return events[0]
        return EventSequence(*events)


    @abstractmethod
    def _create_from_str(cls_or_self, s, **kwargs):
        raise NotImplementedError





class StringEvent(Event, StringAnalyzeUtilitites, str):

    #__slots__ = ["name"]
    # memory saving usually negligible and inheritance from non slot class is
    # impossible
    
    allowed_names = () # set allowed_names even without use of NamedClass
    

    def __new__(cls, name="", *, only_defined_names=False):
        if isinstance(name,int): name = str(name)
        if name != "":
            if cls.allowed_names and only_defined_names:
                if name not in cls.allowed_names:
                    raise NameError(f"name '{name}' not allowed for {cls}")
                #if name.isnumeric(): name = f"[{name}]"
        self = str.__new__(cls, name)
        self.name = name
        return self
    
    @classmethod
    def _args_from_str(cls, s):
        # OVERRIDE THIS!
        if s.startswith("[") and s.endswith("]"):
            s = s[1:-1]
        return (s,)
    
    
    @classmethod
    def _create_from_str(cls, s, **kwargs):
        # don't override
        return cls(*cls._args_from_str(s),**kwargs)
    
    create_from_str = classmethod(StringAnalyzeUtilitites.create_from_str)



    
    
    
class NamedEvent(StringEvent):
    
    _special_attr = ["given_name","named_instance"]
    
    NamedClass = None  # Usually set by __init_subclass__ method of NamedClass
    
    
    def __new__(cls, name="", *, only_defined_names=False):
        given_name = name
        named_instance = None
        if isinstance(name,int): name = str(name)
        if name != "":
            try:
                # print(name, named_instance)
                named_instance = cls.NamedClass.instance(name)
            except KeyError:
                if not cls.allowed_names and only_defined_names:
                    raise NameError(f"name '{name}' not allowed for {cls}")
            else:
                name = named_instance.name
        self = super().__new__(cls, name, only_defined_names=only_defined_names)
        self.given_name = given_name
        self.named_instance = named_instance
        return self

    # problem with redefining __eq__ and dicts/set __hash__ comparison
    # __hash__ method is needed for checking "self in dict" or "self in set"
    # however, as each name_string of the NamedKey has a unique hash,
    # and the __contains__ method of set and dict is only checking for
    # hash equality,
    # (while __eq__-equality is checked only after hash equality)
    # we would need to provide several hash values for each keyvent
    # object, which is impossible
    # workarounds:
    # 1. self in NamedKey.standardize(dict/set)
    #    this only checks for stnd name of self.
    #    the search inside the dict is using hashes and thus is quick
    #    however, standardization needs to be done beforehand
    # 2. custom methods defined below.

    def eq(self, other):
        if isinstance(other, int): other = f"[{other}]"
        if isinstance(other, str):
            if self.named_instance:
                return self.named_instance == other
            else:
                return super().__eq__(other)
        raise NotImplementedError

    def isin(self, *others):
        if len(others) == 0:
            raise SyntaxError
        elif len(others) == 1:
            others = others[0]
        for item in iter(others):
            if self.eq(item): return True
        return False

    def get_from(self,dic,return_if_not_found=None):
        try:
            c=dic.NamedClass_for_this_dict
            # this attribute only exists if dic is of type StandardizingDictClass
        except AttributeError:
            pass
        else:
            if c == self.NamedClass:
                # in this case, the dic is already standardized,
                # so we can use a normal hash comparison
                return dic.get(self, return_if_not_found)
        # this is what usually happens
        for key in iter(dic):
            if self.eq(key): return dic[key]
        return return_if_not_found


class PressReleaseMixin:
    
    
    def __new__(cls, name="", press=True, *, write_rls=True,
            only_defined_names=False):
        assert press in (True, False, None)
        self = super().__new__(cls, name, only_defined_names=only_defined_names)
        if press is False and write_rls:
            self2 = str.__new__(cls, self.name + "_rls")
            self2.name = self.name
            try:
                special_attr = cls._special_attr
            except AttributeError:
                pass
            else:
                for attr in special_attr:
                    setattr(self2,attr,getattr(self,attr))
            self = self2
        self.press=press
        return self

    @classmethod
    def _args_from_str(cls, s):
        rls = s.endswith("_rls")
        if rls: s = s[:-4]
        return super()._args_from_str(s) + (not rls,)
    
    def reverse(self):
        return self.__class__(name=self.name, press=not self.press)

    def strip_rls(self, raise_error=True):
        if not self: return self  # happrnd for empty string ""
        if raise_error and self.press: raise ValueError
        return self._rls_stripped()
    
    def _rls_stripped(self):
        return self.__class__(name=self.name, press=False, write_rls=False)


class PressReleaseEvent(PressReleaseMixin, StringEvent):
    pass
    #__slots__ = ["press"]

@PressReleaseEvent.register
class NamedPressReleaseEvent(PressReleaseMixin, NamedEvent):
    #__slots__ = ["press"]
    
    def reverse(self):
        try:
            if self.press:
                return self.named_instance.release_event
            else:
                return self.named_instance.press_event
        except AttributeError:
            return super().reverse()
        
    def _rls_stripped(self):
        try:
            return self.named_instance.release_event_without_rls
        except AttributeError:
            return super()._rls_stripped()


class Keyvent(NamedPressReleaseEvent):
    pass
    #__slots__ = []


class Buttonevent(NamedPressReleaseEvent):
    
    #__slots__ = ["x", "y"]
    
    def __new__(cls, *args, x=None, y=None,**kwargs):
        self= super().__new__(cls,*args,**kwargs)
        self.x=x
        self.y=y
        return self









class StringAnalyzer(StringAnalyzeUtilitites):
    
    #inherits method create_from_str!
    
    def __init__(self, *event_classes):
        self.event_classes = []
        for cls in event_classes:
            if issubclass(cls, NamedObj): cls = cls.Event
            if not issubclass(cls, StringEvent): raise TypeError
            if cls is EventSequence: raise TypeError
            self.event_classes.append(cls)
    
    

    def _create_from_str(self, s, only_defined_names=False):
        for cls in self.event_classes:
            try:
                return cls._create_from_str(s, only_defined_names=True)
            except NameError:
                continue
        if only_defined_names:
            raise NameError(f"String '{s}' could not be matched to any "
            f"event class in {self.event_classes} of {self}.")
        else:
            return self.event_classes[0]._create_from_str(s,
                    only_defined_names=False)



class EventObjectSender(ABC):
    
    default_delay=None
    
    @abstractmethod
    def _send_event(self, event, **kwargs):
        raise NotImplementedError
    
    @hotkeys.add_pause_option(True)
    def send_event(self, *events, delay=None,**kwargs):
        delay = self.default_delay if delay is None else delay
        for event in events:
            if isinstance(event, EventSequence):
                self.send_event(*event.members, **kwargs)
            elif isinstance(event, EventCombination):
                self.send_event(event.sending_version(), **kwargs)
            else:
                try:
                    name = event.name
                except AttributeError:
                    pass
                else:
                    if name.startswith("[") and name.endswith("]"):
                        name = name[1:-1]
                        event.name = name
                try:
                    self._send_event(event, **kwargs)
                except TypeError:
                    raise TypeError(f"event argument {event} not allowed for "
                                    f"send_event method of object {self}.")
                if delay: sleep(delay/1000)


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from ... import adaptionmethod, hotkeys, MouseMoveEvent, NamedButton
from ..event_sender import AdaptivePressReleaseSender
from time import sleep

class MouseAdaptor(AdaptivePressReleaseSender):
    
    # inherits adaptionmethods _press and _rls

    default_delay = 0
    """
    .. exec::

        event_sender.default_delay_doc("buttons")
    """

    default_duration = 1
    """
    .. exec::

        event_sender.default_duration_doc("button")
    """
    
    @property
    def NamedClass(self):
        return self.NamedButtonClass
    @NamedClass.setter
    def NamedClass(self, val):
        if not issubclass(val, NamedButton): raise TypeError
        self.NamedButtonClass = val
        self.create_effective_dict()
    
    @property
    def button(self):
        return self.NamedButtonClass.NameContainer
    
    @adaptionmethod
    def pos(self):
        return self.pos.target()
    
    @adaptionmethod
    def move(self, dx, dy):
        return self.move.target_with_args()
    
    @adaptionmethod
    def moveto(self, x=None, y=None):
        if x is None: x = self.pos()[0]
        if y is None: y = self.pos()[1]
        return self.moveto.target(x, y)

    @hotkeys.add_pause_option(True)
    def press(self, button, *buttons, x=None, y=None, delay=None):
        if x or y: self.moveto(x, y)
        return super().press(button, *buttons,delay=delay)

    @hotkeys.add_pause_option(True)
    def rls(self, button, *buttons, x=None, y=None, delay=None):
        if x or y: self.moveto(x, y)
        return super().rls(button, *buttons, delay=delay)


    @adaptionmethod
    @hotkeys.add_pause_option(True)
    def scroll(self, vertical, horizontal=0):
        # it is necessary to explicitely pass the parameters here, because
        # hotkeys.add_pause_option is adding a parameter which confuses
        # call_target
        return self.scroll.target(vertical, horizontal)

    @hotkeys.add_pause_option(True)
    def click(self, button=1, count=1, x=None, y=None, duration=None,
            delay=None):
        if x or y: self.moveto(x, y)
        self.tap(button,delay=delay, duration=duration, repeat=count)
        
    
    def rclick(self, *args, **kwargs):
        return self.click("right", *args, **kwargs)
    
    def mclick(self, *args, **kwargs):
        return self.click("middle", *args, **kwargs)
    
    def click2(self, button=1, *args, **kwargs):
        return self.click(button, 2, *args, **kwargs)
    
    def click3(self, button=1, *args, **kwargs):
        return self.click(button, 3, *args, **kwargs)
    
    def clickrel(self, button=1, count=1, dx=0, dy=0, **kwargs):
        self.move(dx, dy)
        self.click(button, count, **kwargs)
    
    @hotkeys.add_pause_option(True)
    def drag(self, dx, dy, button=1, x=None, y=None):
        self.press(button, x=x, y=y)
        self.move(dx, dy)
        self.rls(button)
    
    @hotkeys.add_pause_option(True)
    def dragto(self, x, y, button=1, x0=None, y0=None):
        self.press(button, x=x0, y=y0)
        self.moveto(x, y)
        self.rls(button)
    
    
    def _send_event(self, event, delay=None, reverse_press=False,
            autorelease=False):
        try:
            return super()._send_event(event, delay, reverse_press,
                autorelease)
        except TypeError:
            if isinstance(event, MouseMoveEvent):
                if event.relative:
                    self.move(event.x, event.y)
                else:
                    self.moveto(event.x, event.y)
            else:
                raise TypeError


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .. import Adaptor, adaptionmethod


class NotificationAdaptor(Adaptor):
    
    active = True
        
    def __call__(self, title: str, timeout=3, text: str = ""):
        if self.active:
            title = str(title)
            text=str(text)
            if title == "": title = " "
            self.post_notification(title,timeout,text)
            return True
        return False
        
    @adaptionmethod
    def post_notification(self, title, timeout, text):
        return self.post_notification.target_with_args()
    
    def activate(self):
        self.active = True
    
    def deactivate(self):
        self.active = False
    
    def toggle(self):
        self.__call__(f"Notifications turned off.")
        self.active = not self.active
        self.__call__(f"Notifications turned on.")
        
    def toggle_all(self):
        for i in self.get_instances(): i.toggle()
    


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .keybutton_classes import NamedKey, NamedButton



@NamedButton.update_names
class StandardButtonNames:
    class standard:
        left = 'mouse_left', 'mleft', 'left', 'l', 'mouse_1', 'm1', 1
        middle = 'mouse_middle', 'mmiddle', 'middle', 'm', 'mouse_2', 'm2', 2
        right = 'mouse_right', 'mright', 'r', 'right', 'mouse_3', 'm3', 3
    class scrolling:
        scroll_up = 'scroll_up', 's_up', 'wheel_up', 'mouse_4', 'm4', 4
        scroll_down = 'scroll_down', 's_down', 'wheel_down', 'mouse_5', 'm5', 5
        scroll_left = 'scroll_left', 's_left', 'wheel_left', 'mouse_6', 'm6', 6
        scroll_right = 'scroll_right', 's_right', 'wheel_right', 'mouse_7', \
            'm7', 7
    class unnamed:
        mouse_0 = "mouse_0", "m0", 0
    # fill the unnamed class with following buttons names:
    for i in range(8, 40):
        name = f"mouse_{i}"
        setattr(unnamed, name, (name, f"m{i}", i))
    del i, name


@NamedKey.update_names
class StandardKeyNames:
    class small_letters:
        a = 'a'
        b = 'b'
        c = 'c'
        d = 'd'
        e = 'e'
        f = 'f'
        g = 'g'
        h = 'h'
        i = 'i'
        j = 'j'
        k = 'k'
        l = 'l'
        m = 'm'
        n = 'n'
        o = 'o'
        p = 'p'
        q = 'q'
        r = 'r'
        s = 's'
        t = 't'
        u = 'u'
        v = 'v'
        w = 'w'
        x = 'x'
        y = 'y'
        z = 'z'
    class capital_letters:
        A = 'A'
        B = 'B'
        C = 'C'
        D = 'D'
        E = 'E'
        F = 'F'
        G = 'G'
        H = 'H'
        I = 'I'
        J = 'J'
        K = 'K'
        L = 'L'
        M = 'M'
        N = 'N'
        O = 'O'
        P = 'P'
        Q = 'Q'
        R = 'R'
        S = 'S'
        T = 'T'
        U = 'U'
        V = 'V'
        W = 'W'
        X = 'X'
        Y = 'Y'
        Z = 'Z'
    class letters(small_letters, capital_letters):
        pass
    class digits:
        zero = '0', 'zero'
        one = '1', 'one'
        two = '2', 'two'
        three = '3', 'three'
        four = '4', 'four'
        five = '5', 'five'
        six = '6', 'six'
        seven = '7', 'seven'
        eight = '8', 'eight'
        nine = '9', 'nine'
    class alphanumeric(letters,digits):
        pass
    class special_text_keys:
        space = 'space', ' '
        enter = 'Enter', 'Return', 'ret'
        backspace = 'Backspace', 'bs', 'Control_h'
        delete = 'Delete', 'del'
        tab = 'Tab','Control_i'
    class numpad_dead:
        p_page_down = 'p_page_down', 'kp_page_down'
        p_page_up = 'p_page_up', 'kp_page_up'
        p_next = 'p_next', 'kp_next'
        p_begin = 'p_begin', 'kp_begin'
        p_up = 'p_up', 'kp_up'
        p_home = 'p_home', 'kp_home'
        p_subtract = 'p_subtract', 'kp_subtract', 'kpminus'
        p_multiply = 'p_multiply', 'kp_multiply', 'kpasterisk'
        p_add = 'p_add', 'kp_add', 'kpadd', 'kpplus'
        p_end = 'p_end', 'kp_end'
        p_delete = 'p_delete', 'kp_delete'
        p_divide = 'p_divide', 'kp_divide', 'kpslash'
        p_insert = 'p_insert', 'kp_insert', 'p_ins'
        p_right = 'p_right', 'kp_right'
        p_left = 'p_left', 'kp_left'
        p_down = 'p_down', 'kp_down'
        p_dot = 'p_dot', 'kp_dot', 'kpdot', 'kp_dec'
        p_enter = 'p_enter', 'kpenter'
    class numpad_digit:
        p_0 = 'p_0', 'kp_0', 'kp0'
        p_1 = 'p_1', 'kp_1', 'kp1'
        p_2 = 'p_2', 'kp_2', 'kp2'
        p_3 = 'p_3', 'kp_3', 'kp3'
        p_4 = 'p_4', 'kp_4', 'kp4'
        p_5 = 'p_5', 'kp_5', 'kp5'
        p_6 = 'p_6', 'kp_6', 'kp6'
        p_7 = 'p_7', 'kp_7', 'kp7'
        p_8 = 'p_8', 'kp_8', 'kp8'
        p_9 = 'p_9', 'kp_9', 'kp9'
    class numpad(numpad_dead, numpad_digit):
        pass
    class navigation:
        Home = 'Home', 'Find'
        End = 'End', 'Select'
        PageUp = 'PageUp', 'Prior', 'Page_Up'
        PageDown = 'PageDown', 'Next', 'Page_Down'
        Esc = 'Esc', 'Escape'
    class modifiers:
        ShiftR = 'ShiftR', 'Shift_R', 'rightshift'
        ShiftL = 'ShiftL', 'Shift', 'Shift_L', 'leftshift'
        CtrlL = 'CtrlL', 'Control_L', 'Ctrl', 'Ctrl_L', 'Control', 'leftcontrol',\
            'leftctrl', "LCtrl", "L_Ctrl"
        CtrlR = 'CtrlR', 'Control_R', 'Ctrl_R', 'rightcontrol', 'rightctrl', \
            "RCtrl", "R_Ctrl"
        Alt = 'Alt', 'Alt_L', 'AltL', 'leftalt'
        AltGr = 'AltGr', 'Alt_R', 'AltR', 'rightalt', '65027'
        SControl = 'SControl', 'SCtrl'
        Win = 'Win', 'Super', 'Super_L', 'cmd', 'leftmeta'
    class state_lock:
        Num_Lock = 'Num_Lock', 'NLock', 'NumLock'
        Scroll_Lock = 'Scroll_Lock', 'SLock', 'ScrollLock'
        Caps_Lock = 'Caps_Lock', 'CLock', 'CapsLock'
        AltGr_Lock = 'AltGr_Lock', 'AltRLock'
        Alt_Lock = 'Alt_Lock', 'AltLLock', 'AltLock'
    class arrow_keys:
        up = 'up'
        down = 'down'
        left = 'left'
        right = 'right'
    class fkeys:
        F1 = 'F1'
        F2 = 'F2'
        F3 = 'F3'
        F4 = 'F4'
        F5 = 'F5'
        F6 = 'F6'
        F7 = 'F7'
        F8 = 'F8'
        F9 = 'F9'
        F10 = 'F10'
        F11 = 'F11'
        F12 = 'F12'
    class special_functions:
        wlan = 'wlan'
        volumeup = 'volumeup'
        volumedown = 'volumedown'
        play = 'play', 'playpause', "playcd", "player"
        stop = 'stop', 'stopcd'
        nextsong = 'nextsong'
        previoussong = 'previoussong'
        mute = 'mute', 'min_interesting'
        touchpad_on = 'touchpad_on'
        touchpad_off = 'touchpad_off'
        appselect = 'appselect'
        sleep = 'sleep'
        calc = 'calc'
    
    class dead_keys(fkeys, special_functions, arrow_keys,state_lock,
            modifiers,navigation, numpad_dead):
        compose = 'compose'
        Linefeed = 'Linefeed', 'Control_j'
        Menu = 'Menu'
        print = 'Print', 'print_screen', 'sysrq'
        pause = 'Pause'
        insert = 'Insert', 'ins'
        
        
    class symbols:
        exclam = 'exclam', '!'
        quotedbl = 'quotedbl', '"'
        numbersign = 'numbersign', '#'
        dollar = 'dollar', '$'
        percent = 'percent', '%'
        ampersand = 'ampersand', '&'
        apostrophe = 'apostrophe', "'"
        parenleft = 'parenleft', '('
        parenright = 'parenright', ')'
        asterisk = 'asterisk', '*'
        plus = 'plus', '+', 'add'
        comma = 'comma', ','
        minus = 'minus', '-', 'subtract'
        period = 'period', '.', 'dot'
        slash = 'slash', '/', 'divide'
        backslash = 'backslash', '\\'
        colon = 'colon', ':'
        semicolon = 'semicolon', ';'
        less = 'less', '<', '102nd'
        equal = 'equal', '='
        greater = 'greater', '>'
        question = 'question', '?'
        at = 'at', '@'
        bracketleft = 'bracketleft', '[', 'leftbracket'
        bracketright = 'bracketright', ']', 'rightbracket'
        asciicircum = 'asciicircum', '^', 'circumflex'
        underscore = 'underscore', '_'
        grave = 'grave', '`'
        braceleft = 'braceleft', '{', 'leftbrace'
        bar = 'bar', '|'
        braceright = 'braceright', '}', 'rightbrace'
        asciitilde = 'tilde', '~', 'asciitilde'
        exclamdown = 'exclamdown', '¡'
        cent = 'cent', '¢'
        sterling = 'sterling', '£', 'pound'
        currency = 'currency', '¤'
        yen = 'yen', '¥'
        brokenbar = 'brokenbar', '¦'
        section = 'section', '§'
        diaeresis = 'diaeresis', '¨'
        copyright = 'copyright', '©'
        ordfeminine = 'ordfeminine', 'ª'
        guillemotleft = 'guillemotleft', '«'
        notsign = 'notsign', '¬'
        registered = 'registered', '®'
        macron = 'macron', '¯'
        degree = 'degree', '°'
        plusminus = 'plusminus', '±'
        twosuperior = 'twosuperior', '²'
        threesuperior = 'threesuperior', '³'
        acute = 'acute', '´'
        mu = 'mu', 'µ'
        paragraph = 'paragraph', '¶', 'pilcrow'
        periodcentered = 'periodcentered', '·'
        cedilla = 'cedilla', '¸'
        onesuperior = 'onesuperior', '¹'
        masculine = 'masculine', 'º'
        guillemotright = 'guillemotright', '»'
        onequarter = 'onequarter', '¼'
        onehalf = 'onehalf', '½'
        threequarters = 'threequarters', '¾'
        questiondown = 'questiondown', '¿'
        Agrave = 'Agrave', 'À'
        Aacute = 'Aacute', 'Á'
        Acircumflex = 'Acircumflex', 'Â'
        Atilde = 'Atilde', 'Ã'
        Adiaeresis = 'Adiaeresis', 'Ä'
        Aring = 'Aring', 'Å'
        AE = 'AE', 'Æ'
        Ccedilla = 'Ccedilla', 'Ç'
        Egrave = 'Egrave', 'È'
        Eacute = 'Eacute', 'É'
        Ecircumflex = 'Ecircumflex', 'Ê'
        Ediaeresis = 'Ediaeresis', 'Ë'
        Igrave = 'Igrave', 'Ì'
        Iacute = 'Iacute', 'Í'
        Icircumflex = 'Icircumflex', 'Î'
        Idiaeresis = 'Idiaeresis', 'Ï'
        ETH = 'ETH', 'Ð'
        Ntilde = 'Ntilde', 'Ñ'
        Ograve = 'Ograve', 'Ò'
        Oacute = 'Oacute', 'Ó'
        Ocircumflex = 'Ocircumflex', 'Ô'
        Otilde = 'Otilde', 'Õ'
        Odiaeresis = 'Odiaeresis', 'Ö'
        multiply = 'multiply', '×', 'multiplication'
        Ooblique = 'Ooblique', 'Ø', 'Oslash'
        Ugrave = 'Ugrave', 'Ù'
        Uacute = 'Uacute', 'Ú'
        Ucircumflex = 'Ucircumflex', 'Û'
        Udiaeresis = 'Udiaeresis', 'Ü'
        Yacute = 'Yacute', 'Ý'
        THORN = 'THORN', 'Þ'
        ssharp = 'ssharp', 'ß'
        agrave = 'agrave', 'à'
        aacute = 'aacute', 'á'
        acircumflex = 'acircumflex', 'â'
        atilde = 'atilde', 'ã'
        adiaeresis = 'adiaeresis', 'ä'
        aring = 'aring', 'å'
        ae = 'ae', 'æ'
        ccedilla = 'ccedilla', 'ç'
        egrave = 'egrave', 'è'
        eacute = 'eacute', 'é'
        ecircumflex = 'ecircumflex', 'ê'
        ediaeresis = 'ediaeresis', 'ë'
        igrave = 'igrave', 'ì'
        iacute = 'iacute', 'í'
        icircumflex = 'icircumflex', 'î'
        idiaeresis = 'idiaeresis', 'ï'
        eth = 'eth', 'ð'
        ntilde = 'ntilde', 'ñ'
        ograve = 'ograve', 'ò'
        oacute = 'oacute', 'ó'
        ocircumflex = 'ocircumflex', 'ô'
        otilde = 'otilde', 'õ'
        odiaeresis = 'odiaeresis', 'ö'
        division = 'division', '÷'
        oslash = 'oslash', 'ø'
        ugrave = 'ugrave', 'ù'
        uacute = 'uacute', 'ú'
        ucircumflex = 'ucircumflex', 'û'
        udiaeresis = 'udiaeresis', 'ü'
        yacute = 'yacute', 'ý'
        thorn = 'thorn', 'þ'
        ydiaeresis = 'ydiaeresis', 'ÿ'


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .. import DependencyManager

with DependencyManager(__name__) as manager:
    pynput = manager.import_module("pynput")

from .baseclasses import (InputEventHandler, KeyhookBase, ButtonhookBase,
    CursorhookBase)

keyboard = pynput.keyboard.Controller()


class PynputHandler(InputEventHandler):
    
    
    collector = None
    capturer = None
    
    def start_collecting(self):
        self.collector = self.get_collector()
        self.collector.start()
    def stop_collecting(self):
        self.collector.stop()
        
    def start_capturing(self):
        self.capturer = self.get_capturer()
        self.capturer.start()
    def stop_capturing(self):
        self.capturer.stop()
        

    def get_collector(self):
        raise NotImplementedError
    

    def get_capturer(self):
        raise NotImplementedError


class PynputKeyHandler(PynputHandler):
    
    
    def get_collector(self):
        return pynput.keyboard.Listener(on_press=self.key_down_event,
                on_release=self.key_up_event)
    
    def key_down_event(self, event):
        name = self._get_key_info(event)
        self.queue_event(name, press=True)
    
    def key_up_event(self, event):
        name = self._get_key_info(event)
        self.queue_event(name, press=False)

    @staticmethod
    def _get_key_info(key_object):
        if type(key_object) is pynput.keyboard.Key:
            return key_object.name#, key_object.value.vk
        elif type(key_object) is pynput.keyboard.KeyCode:
            return key_object.char#, key_object.vk
        else:
            raise TypeError

    @staticmethod
    def get_capturer():
        return pynput.keyboard.Listener(suppress=True)


class PynputMouseHandler(PynputHandler):
    def get_collector(self):
        return pynput.mouse.Listener(on_click=self.click_event,
                on_scroll=self.scroll_event)

    def click_event(self, x, y, button, pressed):
        self.queue_event(button.value, press=pressed, x=x, y=y)
    
    def scroll_event(self, x, y, dx, dy):
        d = dx, dy
        if d == (0, 1):
            button = 4  # "scroll_up"
        elif d == (0, -1):
            button = 5  # "scroll_down"
        elif d == (1, 0):
            button = 7  # "scroll_right"
        elif d == (-1, 0):
            button = 6  # "scroll_left"
        else:
            raise ValueError
        self.queue_event(button, press=None, x=x, y=y)

    @staticmethod
    def get_capturer():
        return pynput.mouse.Listener(suppress=True)



class PynputCursorHandler(PynputHandler):
    
    capture_allowed = False
    
    def get_collector(self):
        return pynput.mouse.Listener(on_move=self.queue_event)
    



class Keyhook(KeyhookBase):
    handler = PynputKeyHandler()

class Buttonhook(ButtonhookBase):
    handler = PynputMouseHandler()

class Cursorhook(CursorhookBase):
    handler = PynputCursorHandler()


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from . import HookAdaptor, CallbackHook, hotkeys, KeyboardAdaptor, \
    MouseAdaptor
from Dhelpers.arghandling import check_type
from Dhelpers.baseclasses import TimedObject
from Dhelpers.launcher import launch
import time
from .. import NotificationAdaptor



class Waiter(TimedObject):
    ntfy = NotificationAdaptor(group="waiter", _primary_name="ntfy")
    
    def __init__(self, callback_hook=None, maxlen = None, maxtime = 20, *,
            endevents=(), cond_func=None, eventmap=None,
            wait=True, capture=False, notify=False, timestamps=False):
        super().__init__(timeout=maxtime, wait=wait)
        check_type(CallbackHook, callback_hook)
        if callback_hook: raise ValueError(f"Arguments of callback hook "
            f"{callback_hook} were already set.")
        self.maxlen = maxlen
        self.maxtime = maxtime
        if type(endevents) not in (list, tuple): endevents = [endevents]
        self.endevents =endevents
        self.eventmap = eventmap
        self.wait = wait
        self.capture = capture
        if cond_func and not callable(cond_func):
            raise ValueError(f"KeyWaiter cond_func {cond_func} is not callable")
        self.cond_func = cond_func
        timeout = maxtime + 5 if maxtime else None
        self.callback_hook = callback_hook(self.called, timeout,
                capture=self.capture)
        self.num = 0
        self.events = []
        self.timestamps = timestamps
        self.event_timestamps=[]
        if self.eventmap: self.events_mapped = []
        self.notify = notify
    
    def _start_action(self):
        if self.notify:  self.ntfy("Waiter active", 0.5,
                        f"maxlen:{self.maxlen}, maxtime:{self.maxtime}")
        self.callback_hook.start()
        time.sleep(0.1)
        # this time is needed for the hook manager to initialize
    
    def _stop_action(self):
        self.callback_hook.stop()

    def called(self, event):
        if self.timestamps: self.event_timestamps.append(time.time())
        self.events.append(event)
        self.num += 1
        event_mapped = None
        if self.eventmap:
            try:
                event_mapped = self.eventmap[event]
            except KeyError:
                # if no entry found, omit it and do not count event
                self.num -= 1
            else:
                self.events_mapped.append(event_mapped)
        exitcode = self.event_condition(event, event_mapped)
        if exitcode is not None: self.stop(exitcode)
    
    
    def event_condition(self, event, ev_mapped):
        if self.maxlen and self.num >= self.maxlen:
            return "maxlen"
        if event in self.endevents or ev_mapped in self.endevents:
            return "endevent"
        f = self.cond_func
        if f:
            try:
                argcount = f.__code__.co_argcount
                if argcount == 0:
                    return f()
                elif argcount == 1:
                    return f(self)
                elif argcount == 2:
                    return f(self, event)
                elif argcount == 3:
                    return f(self, event, ev_mapped)
                else:
                    raise SyntaxError
            except Exception as e:
                self.stop(e)
                raise
            
    def event_intervalls(self):
        if not self.timestamps: raise ValueError
        before = self.starttime
        intervalls = []
        for t in self.event_timestamps:
            intervalls.append(t-before)
            before = t
        return intervalls

    
    @classmethod
    def get1(cls, callback_hook, maxtime =2, wait = True, **options):
        if wait is not True:
            raise Exception("ERROR: get1 a arg 'wait' must be true")
        get1 = cls(callback_hook, maxlen=1, maxtime = maxtime, wait=True,
                **options).start()
        if get1.exitcode == "maxlen":
            return get1.events[0]
        if get1.exitcode != "timeout": raise RuntimeError
        
    # def reinject(self, delay=10, autorelease=False, reverse=False):
    #     delay = delay/1000
    #     if self.exitcode == "active": raise ValueError
    #     for event in self.events:
    #         event.reinject(autorelease, reverse=reverse)
    #         time.sleep(delay)
        
        
class KeyWaiter(Waiter, HookAdaptor.AdaptiveClass):
    
    adaptor = HookAdaptor(group="keywait", _primary_name="KeyWaiter.adaptor")
    hook = adaptor
    keyb = KeyboardAdaptor(group="keywait",_primary_name="keyb")
    mouse = MouseAdaptor(group="keywait", _primary_name="mouse")
    hotstring_keyb = KeyboardAdaptor(group="hotstring")

    def __init__(self, *args, eventmap=None, keys=True,  buttons=False,
            press=True, release=False, write_rls=True, join_events=False, **kwargs):
        
        if release and write_rls and eventmap:
            eventmap.update({i + "_rls": j + "_rls" for i, j in eventmap.items()})
        if not release and not press:
            raise ValueError("KeyWaiter: key ups and down events are both disabled")
        self.join_events = join_events
        if join_events:
            self.joined_events = ""
            self.joined_events_mapped = ""
            
        hook_creator = sender = 0
        if keys:
            hook_creator += self.hook.keys()
            sender += self.keyb
        if buttons:
            hook_creator += self.hook.buttons()
            sender += self.mouse
        callback_hook = hook_creator(press=press, release=release,
                write_rls=write_rls)
        self.sender = sender
        self.press = press
        self.release = release
        self.write_rls = write_rls
        self.keys = keys
        self.buttons = buttons
        super().__init__(callback_hook, *args, eventmap=eventmap, **kwargs)
    
    def event_condition(self, event, ev_mapped):
        #print(event)
        if self.join_events:
            self.joined_events += event
            if ev_mapped: self.joined_events_mapped += ev_mapped
        return super().event_condition(event, ev_mapped)
    
    
    def reinject(self,delay=50):
        #_print(self.events, self.keyb.name_translation)
        reverse_press = False
        if self.press and self.release: autorelease = False
        elif self.press and not self.release: autorelease = True
        elif not self.press and self.release:
            autorelease = True
            reverse_press = True
        else: raise ValueError
        self.sender.send_event(*self.events, delay=delay,
                autorelease =autorelease, reverse_press= reverse_press)

    @classmethod
    def get1key(cls, maxtime = 2, wait = True, press=True, release=True,
            capture=True, notify=True, **options):
        if wait is not True:
            raise Exception("ERROR: get1 keyword arg 'wait' must be true")
        with hotkeys.paused(3):
            get1 = cls(maxlen=1, maxtime=maxtime, wait=True, press=press,
                release = release, capture=capture, notify=notify, **options)
            get1.start()
        if get1.exitcode == "maxlen":
            key = get1.events[0]
            if not press and not key.press and not options.get("write_rls",
                    False):
                # if only release events are collected, and if write_rls was
                # not explicitly set to True, remove the _rls tag by default
                key = key.strip_rls()
        elif get1.exitcode == "timeout":
            key = cls.hook.NamedKeyClass.Event() #return empty event
        else:
            raise RuntimeError
        return key


    @classmethod
    def hotstring(cls, string_dict, send=True, undo=True, undo_additional=0,
            endevents = (), limit_allowed_keys: list = False, eventmap=None):
        
        if limit_allowed_keys:
            limit_allowed_keys += [i + "_rls" for i in limit_allowed_keys]
        elif endevents == ():
            endevents = ("Tab", "Return", "Space", "Backspace")
        patterns = tuple(string_dict.keys())
        def checkinp(self, k, k_mapped):
            if limit_allowed_keys:
                if self.eventmap and k_mapped not in limit_allowed_keys:
                    return f"forbidden key {k_mapped}"
                if not self.eventmap and k not in limit_allowed_keys:
                    return f"forbidden key {k}"
            y = self.joined_events_mapped if self.eventmap else self.joined_events
            if y in patterns: return "hit"
    
        with hotkeys.paused(15):
            inp = cls(20, 5, endevents=endevents,eventmap=eventmap,
                    cond_func=checkinp, press=False, release=True,
                    write_rls=False, capture=False, join_events=True).start()
        if inp.exitcode == "hit":
            if undo:
                cls.hotstring_keyb.send("<BackSpace>"*(inp.num +
                                                      undo_additional))
            if send:
                if eventmap:
                    cls.hotstring_keyb.send(string_dict[
                        inp.joined_events_mapped])
                else:
                    cls.hotstring_keyb.send(string_dict[inp.joined_events])
        return inp
    
    @classmethod
    def wait_for_key(cls, *keynames, timeout=20):
        inp = cls(maxtime=timeout, endevents=keynames, press=False,
                release=True, write_rls=False, wait=True)
        inp.start()
        if inp.exitcode == "timeout": return False
        return inp.duration()


#
#
# Copyright (c) 2020-2025 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import shlex, re
from Dhelpers.arghandling import CollectionWithProps, NonNegativeInt

from .. import DependencyManager, launch

with DependencyManager(__name__) as manager:
    current_id = manager.test_shellcommand("xdotool",test_cmd ="xdotool "
                                   "getactivewindow", pkg="xdotool")
    #checks if xdotool is installed and if getactivewindow works

    manager.test_shellcommand("wmctrl", pkg="wmctrl")

    manager.test_shellcommand("xprop", test_cmd=f"xprop -id {current_id} " \
                                           "WM_CLASS")
    # checks if xprop is installed and if it gets the current window's class



def screen_res():
    command = "xdotool getdisplaygeometry"
    a, b = launch.get(command).split(" ")
    return int(a), int(b)


def _xdo_cmd(title=None, cl=None, cl_name=None, pid=None, ID=None,
        all_param=True, onlyvisible=False, limit=0, add_final=None):
    cmd = ["xdotool"]
    if ID:
        cmd += shlex.split(add_final) + [ID]
    else:
        if title == ":ACTIVE:":
            cmd += ["getactivewindow"]
        elif title == ":SELECT:":
            cmd += ["selectwindow"]
        else:
            cmd += ["search"]
            if onlyvisible:
                cmd += ["--onlyvisible"]
            if limit > 0:
                cmd += ["--limit", limit]
            if all_param:
                cmd += ["--all"]
            if cl_name:
                cmd += ["--classname", cl_name]
            if cl:
                cmd += ["--class", cl]
            if pid:
                cmd += ["--pid", pid]
            if title:
                cmd += ["--name", "--", title]
                # the additional -- are important to make clear that title
                # can start with - and is still not a parameter
        if add_final:
            cmd += shlex.split(add_final)
    # print(cmd)
    return cmd

# sometimes executing xdotool results in a X Error of failed request:
# BadWindow (invalid Window parameter). (
# Somehow always when I use ntfy)
# In this case xdotool does not find a window although it exists. So we will
#  use a cheap workaround:


# def _bash_wait_workaround(command, **unimportant):
#     # if the window exists, but is not found due to an X Error, we can try up
#     #  to three times.
#     for j in range(3):
#         x = launch.get(command, check=False, check_stderr=False)
#         if x:
#             return x
#         if j > 2:
#             time.sleep(0.01)
#     print(x)
#     return x



def _xdo_get_from_single_id(ID, add_final):
    # add_final contains the information about which information to retrieve
    # from window with given ID
    x = launch.get(_xdo_cmd(ID=ID, add_final=add_final), check=False,
            check_err=False)
    if x: return x

# find all window ids matching the given pattern

def _xdo_get_ids(*win_args, **win_kwargs):
    cmd = _xdo_cmd(*win_args,  **win_kwargs, add_final=None)
        # add_final=None makes sure that IDs are returned
    try:
        x = launch.get(cmd)
    except launch.CalledProcessError:
        return set()
    if type(x) is str:
        return [int(i) for i in x.split("\n")]



def ID_from_location(param):
    if isinstance(param,
            CollectionWithProps(NonNegativeInt, len=2, allow_set=False)):
        x, y = param
        command = ["xdotool", "mousemove", x, y, "getmouselocation",
            "mousemove", "restore"]
        return int(re.split('[\\s:]', launch.get(command))[7])
    elif isinstance(param, str):
        param = param.lower()
        if param in ("m", "mouse"):
            command = ["xdotool", "getmouselocation"]
            return int(re.split('[\\s:]', launch.get(command))[7])
        elif param in ("a", "act", "active"):
            return _xdo_get_ids(title=":ACTIVE:")[0]
        elif param in ("s", "sel", "select"):
            return _xdo_get_ids(title=":SELECT:")[0]
    raise NotImplementedError

allowed_properties = "title", "wcls", "pid"

def IDs_from_property(prop_name, prop_val, visible):
    if prop_name == "title":
        return _xdo_get_ids(title=prop_val, onlyvisible=visible)
    elif prop_name == "wcls":
        wcls = prop_val
        if "." in wcls:
            cl_name, cl = wcls.split(".")
            # cls is a string of pattern "cl_name.cl" usually
            return tuple(set(_xdo_get_ids(cl=cl, onlyvisible=visible)) & set(
                    _xdo_get_ids(cl_name=cl_name, onlyvisible=visible)))
        else:
            return _xdo_get_ids(cl=wcls, onlyvisible=visible)
    elif prop_name == "pid":
        return _xdo_get_ids(pid=prop_val, onlyvisible=visible)
    raise NotImplementedError


def property_from_ID(ID, prop_name):
    if prop_name == "title":
        return _xdo_get_from_single_id(ID, "getwindowname")
    elif prop_name == "wcls":
        y = launch.get("xprop", "-id", ID, "WM_CLASS")
        x = re.split('"', y)
        if len(x) < 4:
            if y == "WM_CLASS:  not found.": return None
            return y
        return x[-4] + '.' + x[-2]
    elif prop_name == "pid":
        p = _xdo_get_from_single_id(ID, "getwindowpid")
        if p is None: return
        return int(p)
    elif prop_name == "geometry":
        g = _xdo_get_from_single_id(ID, "getwindowgeometry -shell")
        x = re.split("[\n=”]", g)
        return int(x[3]), int(x[5]), int(x[7]), int(x[9])
        # pos x, posy, width, height
    raise NotImplementedError


def activate(ID):
    return _xdo_get_from_single_id(ID, "windowactivate --sync")

def map(ID):
    return _xdo_get_from_single_id(ID,"windowmap --sync")

def unmap(ID):
    return _xdo_get_from_single_id(ID,"windowunmap --sync")

def set_prop(ID, action: str, prop: str, prop2):
    cmd = ["wmctrl", "-i", "-r", ID, "-b", action + "," + prop]
    if prop2:
        cmd[-1] += "," + prop2
    return launch.get(cmd)


def move(ID, x, y, width, height):
    # print("wmctrl -r -i %s -e '0,%s,%s,%s,%s'" % (winID,x,y,width,height))
    return launch.get("wmctrl -i -r %s -e '0,%s,%s,%s,%s'"%(
        ID, int(x), int(y), int(width), int(height)))


def close(ID):
    return _xdo_get_from_single_id(ID, "windowclose")


def kill(ID):
    return _xdo_get_from_single_id(ID, "windowkill")


def minimize(ID):
    return _xdo_get_from_single_id(ID, "windowminimize")
